@@ -69,9 +69,11 @@ def lint(session):
         BLACK_SPEC,
         ISORT_SPEC,
         "flake8>=4,<5",
+        "flake8-future-annotations",
         "flake8-mutable",
         "flake8-quotes",
         "flake8-tidy-imports",
+        "packaging",
     )
     session.run("python", "tools/check_permissions.py")
     session.run("python", "tools/check_whitespace.py")
@@ -104,6 +106,7 @@ def mypy(session):
 
     session.install("mypy")
     session.run("python", "-m", "mypy", "-p", "acconeer.exptool.a121")
+    session.run("python", "-m", "mypy", "-p", "acconeer.exptool.app.new")
     session.run("python", "-m", "mypy", "tests/unit/a121")
 
 
