@@ -2,6 +2,7 @@ from .entities import (
     INT_16_COMPLEX,
     PRF,
     ClientInfo,
+    Criticality,
     IdleState,
     Metadata,
     PersistentRecord,
@@ -14,14 +15,21 @@ from .entities import (
     SessionConfig,
     StackedResults,
     SubsweepConfig,
+    ValidationError,
+    ValidationResult,
+    ValidationWarning,
 )
 from .mediators import ClientError, Recorder
 from .peripherals import (
+    _H5PY_STR_DTYPE,
     Client,
     H5Record,
     H5Recorder,
+    InMemoryRecord,
+    ServerError,
     load_record,
     open_record,
     save_record,
     save_record_to_h5,
 )
+from .utils import iterate_extended_structure
